import shutil
from pathlib import Path

addon_name = "MyAddon"


classic_link = Path(
    Rf"C:\Program Files (x86)\World of Warcraft\_classic_\Interface\AddOns\{addon_name}"
)
if classic_link.exists():
    classic_link.unlink()
    print(f"deleted {classic_link}")

retail_link = Path(
    Rf"C:\Program Files (x86)\World of Warcraft\_retail_\Interface\AddOns\{addon_name}"
)
if retail_link.exists():
    retail_link.unlink()
    print(f"deleted {retail_link}")

addon_dir = Path() / addon_name
if addon_dir.exists():
    shutil.rmtree(addon_dir)
    print(f"deleted {addon_dir}")



import os
from collections.abc import Iterator
from pathlib import Path

import pytest
import respx
from freezegun import freeze_time as _freeze_time
from respx.router import MockRouter

from tests.fixture.curseforge import setup_mock_cf_api
from tests.fixture.fsenv import FSEnv
from tests.fixture.time import TEST_TIME


@pytest.fixture(autouse=True)
def ch_tmp_dir(tmp_path: Path) -> Iterator[None]:
    restore_cwd = Path.cwd()
    os.chdir(tmp_path)
    yield
    os.chdir(restore_cwd)


@pytest.fixture(autouse=True)
def freeze_time() -> Iterator[None]:
    with _freeze_time(TEST_TIME.datetime):
        yield


@pytest.fixture
def fs_env(tmp_path: Path) -> Iterator[FSEnv]:
    yield FSEnv(root=tmp_path)


@pytest.fixture(autouse=True)
def cf_api_respx() -> Iterator[MockRouter]:
    with respx.mock(assert_all_called=False) as respx_mock:
        yield setup_mock_cf_api(respx_mock)



import importlib.resources
import json
import shutil
from pathlib import Path
from typing import Any, Optional

from attr import frozen

import tests.fixture


def _get_fixture_file_path(subpath: Path) -> Path:
    return Path(
        str(importlib.resources.files(tests.fixture).joinpath(Path("data") / subpath))  # type: ignore
    )


def copy_path(src: Path, dst: Path) -> None:
    if src.is_file():
        shutil.copyfile(src, dst)
    elif src.is_dir():
        shutil.copytree(src, dst, dirs_exist_ok=True)
    else:
        raise Exception(f"{src} is not a file or directory")


@frozen(kw_only=True)
class FSEnv:
    root: Path

    # def place_config(
    #     self, config_fixture_file_name: str, target_name: Optional[str] = None
    # ) -> Path:
    #     return self._place_fixture_path(
    #         data_path=Path("configs") / config_fixture_file_name,
    #         target_name=target_name or "wap.json",
    #     )

    def place_addon(
        self, addon_fixture_dir_name: str, target_name: Optional[str] = None
    ) -> Path:
        return self._place_fixture_path(
            data_path=Path("addons") / addon_fixture_dir_name,
            target_name=target_name or "Addon",
        )

    def place_output_dir(
        self, output_dir_name: str, target_name: Optional[str] = None
    ) -> Path:
        return self._place_fixture_path(
            data_path=Path("output_dirs") / output_dir_name,
            target_name=target_name or "dist/",
        )

    def _place_fixture_path(self, data_path: Path, target_name: str) -> Path:
        full_target_path = self.root / target_name
        source_path = _get_fixture_file_path(data_path)
        copy_path(source_path, full_target_path)

        return full_target_path

    def place_file(
        self,
        target_name: str | Path,
        parents: bool = False,
        exist_ok: bool = False,
        text: str = "",
    ) -> Path:
        path = self.root / target_name
        if parents:
            path.parent.mkdir(parents=True, exist_ok=exist_ok)
        path.write_text(text)
        return path

    def place_dir(
        self, target_name: str, parents: bool = False, exist_ok: bool = False
    ) -> Path:
        path = self.root / target_name
        path.mkdir(parents=parents, exist_ok=exist_ok)
        return path

    def get_config_json(
        self,
        config_fixture_file_name: str,
    ) -> Any:
        config_path = _get_fixture_file_path(Path("configs") / config_fixture_file_name)
        text = config_path.read_text()
        return json.loads(text)

    def write_config(
        self,
        config: Any,
        target_name: Optional[str] = None,
    ) -> Path:
        return self.place_file(target_name or "wap.json", text=json.dumps(config))



import webbrowser
from unittest.mock import patch

import pytest

from tests.cmd_util import invoke_help


@pytest.mark.parametrize(
    "subcommand",
    ["build", "validate", "help", "new-config", "new-project", "publish", None],
)
def test_help(subcommand: str | None) -> None:
    with patch("tests.cmd_util.base.webbrowser.open") as webbrowser_open_mock:
        args: list[str] = []
        if subcommand is not None:
            args.append(subcommand)
        result = invoke_help(args)

        assert result.success

        expected_url = "http://t-mart.github.io/wap"
        if subcommand is not None:
            expected_url += f"/commands/{subcommand}"
        webbrowser_open_mock.assert_called_once_with(expected_url)


@pytest.mark.parametrize(
    "subcommand",
    ["build", "validate", "help", "new-config", "new-project", "publish", None],
)
def test_help_no_browser(subcommand: str | None) -> None:
    with patch("tests.cmd_util.base.webbrowser.open", side_effect=webbrowser.Error):
        args: list[str] = []
        if subcommand is not None:
            args.append(subcommand)
        result = invoke_help(args)

        assert result.success

        assert "See documentation at" in result.stderr



from copy import deepcopy
from typing import Any

_BASIC = {
    "$schema": "https://raw.githubusercontent.com/t-mart/wap/master/src/wap/schema/wap.schema.json",  # noqa: E501
    "name": "Package",
    "author": "Me",
    "version": "1.2.3",
    "wowVersions": {"mainline": "9.2.7", "classic": "4.4.0", "vanilla": "1.14.3"},
    "publish": {"curseforge": {"projectId": "1234", "slug": "thepackage"}},
    "package": [
        {
            "path": "./Addon",
            "toc": {
                "tags": {
                    "Title": "The title",
                    "Notes": "Notes for addon",
                },
                "files": ["Main.lua", "Extra.lua"],
            },
            "include": ["./LICENSE"],
        }
    ],
}


def get_basic_config() -> Any:
    return deepcopy(_BASIC)



from __future__ import annotations

import hashlib
import random
import string
import uuid
from io import BytesIO
from pathlib import Path
from typing import BinaryIO
from zipfile import ZipFile

import pytest
from attrs import frozen
from glom import assign, delete  # type: ignore
from respx.router import MockRouter

from tests.cmd_util import invoke_publish
from tests.curseforge_request import CFUploadRequestContent
from tests.fixture.config import get_basic_config
from tests.fixture.curseforge import CURSEFORGE_TOKEN
from tests.fixture.fsenv import FSEnv
from wap.exception import ConfigError, CurseForgeAPIError, PathMissingError

PACKAGE_NAME = get_basic_config()["name"]
PACKAGE_VERSION = get_basic_config()["version"]


@frozen(kw_only=True)
class File:
    """A file in an archive. Contents are represented by their hash only."""

    path: Path
    content_hash: bytes

    @staticmethod
    def _hash(data: bytes) -> bytes:
        return hashlib.new("sha256", data).digest()

    @classmethod
    def from_path(cls, path: Path, root_at: Path) -> File:
        return cls(
            path=path.relative_to(root_at), content_hash=cls._hash(path.read_bytes())
        )

    @classmethod
    def from_zip_path(cls, zipfile: ZipFile, name: str) -> File:
        with zipfile.open(name) as f:
            return cls(path=Path(name), content_hash=cls._hash(f.read()))


@frozen(kw_only=True)
class Archive:
    """
    An archive represents a collection of files, such as a filesystem directory or in a
    zip file.
    """

    files: set[File]

    @classmethod
    def from_dir(cls, dir: Path, root_at: Path) -> Archive:
        return cls(
            files={
                File.from_path(path, root_at)
                for path in dir.rglob("*")
                if path.is_file()
            }
        )

    @classmethod
    def from_zip(cls, source: Path | BinaryIO) -> Archive:
        zip_file = ZipFile(source)
        return cls(
            files={
                File.from_zip_path(zip_file, zip_info.filename)
                for zip_info in zip_file.infolist()
                if not zip_info.is_dir()
            }
        )


def test_publish_normal(fs_env: FSEnv, cf_api_respx: MockRouter) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_output_dir("basic")

    result = invoke_publish(["--curseforge-token", CURSEFORGE_TOKEN])

    assert result.success

    dir_path = Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}")
    zip_path = Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}.zip")

    assert zip_path.is_file()

    dir_archive = Archive.from_dir(dir_path, root_at=dir_path)
    zip_archive = Archive.from_zip(zip_path)
    assert dir_archive == zip_archive

    versions_route = cf_api_respx.routes["versions"]
    assert versions_route
    assert versions_route.called

    upload_file_route = cf_api_respx.routes["upload-file"]
    assert upload_file_route

    req_content = CFUploadRequestContent.from_request(
        upload_file_route.calls[0].request  # type: ignore
    )
    assert (
        req_content.metadata.items()
        >= {
            "changelog": "",
            "changelogType": "text",
            "displayName": f"{PACKAGE_NAME}-{PACKAGE_VERSION}",
            "releaseType": "alpha",
        }.items()
    )
    # order doesn't matter for versions, so test them this way
    assert set(req_content.metadata["gameVersions"]) == {1001, 1002, 1003}

    assert req_content.file_content_type == "application/zip"
    assert req_content.file_name == f"{PACKAGE_NAME}-{PACKAGE_VERSION}.zip"

    request_archive = Archive.from_zip(BytesIO(req_content.file_stream))
    assert request_archive == zip_archive


def test_publish_without_publish_config(fs_env: FSEnv) -> None:
    fs_env.write_config(delete(get_basic_config(), "publish"))

    result = invoke_publish(["--curseforge-token", CURSEFORGE_TOKEN])

    assert isinstance(result.exception, ConfigError)


def test_publish_without_cf_token(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())

    result = invoke_publish()

    assert isinstance(result.exception, SystemExit)


@pytest.mark.parametrize("cl_type", ["markdown", "text", "html"])
def test_publish_changelog_explicit_type(
    fs_env: FSEnv, cf_api_respx: MockRouter, cl_type: str | None
) -> None:
    # this test also tests the case of no changelog file/text being defined.
    config = get_basic_config()
    changelog_path = "./the-changelog"
    assign(config, "publish.curseforge.changelogFile", changelog_path)
    fs_env.place_file(changelog_path)

    assign(config, "publish.curseforge.changelogType", cl_type)
    expected_type = cl_type

    fs_env.write_config(config)
    fs_env.place_output_dir("basic")

    result = invoke_publish(["--curseforge-token", CURSEFORGE_TOKEN])

    assert result.success

    upload_file_route = cf_api_respx.routes["upload-file"]
    assert upload_file_route
    req_content = CFUploadRequestContent.from_request(
        upload_file_route.calls[0].request  # type: ignore
    )
    assert req_content.metadata["changelogType"] == expected_type
    assert req_content.metadata["changelog"] == ""


@pytest.mark.parametrize(
    "cl_file,expected_type",
    [
        ("cl.md", "markdown"),
        ("cl.markdown", "markdown"),
        ("cl.mArKdOwN", "markdown"),
        ("cl.txt", "text"),
        ("cl.TXT", "text"),
        ("cl.html", "html"),
        ("cl.huh", "text"),
    ],
)
def test_publish_auto_changelog_type_determination(
    fs_env: FSEnv, cf_api_respx: MockRouter, cl_file: str, expected_type: str
) -> None:
    config = get_basic_config()
    assign(config, "publish.curseforge.changelogFile", f"./{cl_file}")
    changelog_text = "".join(random.choice(string.ascii_letters) for _ in range(10))
    fs_env.place_file(cl_file, text=changelog_text)

    fs_env.write_config(config)
    fs_env.place_output_dir("basic")

    result = invoke_publish(["--curseforge-token", CURSEFORGE_TOKEN])

    assert result.success

    upload_file_route = cf_api_respx.routes["upload-file"]
    assert upload_file_route
    req_content = CFUploadRequestContent.from_request(
        upload_file_route.calls[0].request  # type: ignore
    )
    assert req_content.metadata["changelogType"] == expected_type
    assert req_content.metadata["changelog"] == changelog_text


def test_publish_different_config_path(fs_env: FSEnv) -> None:
    config_path = "config.json"
    fs_env.write_config(get_basic_config(), "config.json")
    fs_env.place_output_dir("basic")

    result = invoke_publish(
        ["--curseforge-token", CURSEFORGE_TOKEN, "--config-path", config_path]
    )

    assert result.success


@pytest.mark.parametrize("placed_correctly", [True, False])
def test_publish_different_output_path(fs_env: FSEnv, placed_correctly: bool) -> None:
    fs_env.write_config(get_basic_config())
    output_path = "output"
    fs_env.place_output_dir(
        "basic", output_path if placed_correctly else "somewhere-else/"
    )

    result = invoke_publish(
        ["--curseforge-token", CURSEFORGE_TOKEN, "--output-path", output_path]
    )

    if placed_correctly:
        assert result.success
    else:
        assert isinstance(result.exception, PathMissingError)


@pytest.mark.parametrize("cmd_release_type", ["release", "alpha", "beta", None])
@pytest.mark.parametrize("config_release_type", ["release", "alpha", "beta", None])
def test_publish_release_types(
    fs_env: FSEnv,
    cf_api_respx: MockRouter,
    cmd_release_type: str | None,
    config_release_type: str | None,
) -> None:
    config = get_basic_config()
    if config_release_type:
        assign(config, "publish.curseforge.releaseType", config_release_type)
    fs_env.write_config(config)
    fs_env.place_output_dir("basic")

    args = ["--curseforge-token", CURSEFORGE_TOKEN]
    if cmd_release_type:
        args.extend(["--release-type", cmd_release_type])
    result = invoke_publish(args)

    assert result.success

    upload_file_route = cf_api_respx.routes["upload-file"]
    assert upload_file_route

    req_content = CFUploadRequestContent.from_request(
        upload_file_route.calls[0].request  # type: ignore
    )
    expected_release_type = "alpha"
    if config_release_type:
        expected_release_type = config_release_type
    if cmd_release_type:
        expected_release_type = cmd_release_type
    # order doesn't matter for versions, so test them this way
    assert req_content.metadata["releaseType"] == expected_release_type

    if config_release_type is None and cmd_release_type is None:
        assert "No release type specified" in result.stderr


@pytest.mark.parametrize("token", ["malformed", "random"])
def test_publish_bad_auth(fs_env: FSEnv, token: str) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_output_dir("basic")

    result = invoke_publish(
        ["--curseforge-token", token if token != "random" else str(uuid.uuid4())]
    )

    assert isinstance(result.exception, CurseForgeAPIError)


def test_publish_unknown_wow_version(fs_env: FSEnv) -> None:
    fs_env.write_config(assign(get_basic_config(), "wowVersions.mainline", "999.0.0"))
    fs_env.place_output_dir("basic")

    result = invoke_publish(["--curseforge-token", CURSEFORGE_TOKEN])

    assert isinstance(result.exception, CurseForgeAPIError)


@pytest.mark.parametrize("with_slug", [True, False])
def test_publish_slug(fs_env: FSEnv, with_slug: bool) -> None:
    config = get_basic_config()
    if not with_slug:
        delete(config, "publish.curseforge.slug")
    fs_env.write_config(config)
    fs_env.place_output_dir("basic")

    result = invoke_publish(["--curseforge-token", CURSEFORGE_TOKEN])

    assert result.success

    if with_slug:
        assert "Upload available at https:" in result.stderr
    else:
        assert "Uploaded file" in result.stderr



from tests.cmd_util import invoke_validate
from tests.fixture.config import get_basic_config
from tests.fixture.fsenv import FSEnv


def test_valid_validation(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())

    result = invoke_validate()

    assert result.success
    assert result.stdout == "valid"


def test_invalid_validation(fs_env: FSEnv) -> None:
    fs_env.write_config("this aint valid!")

    result = invoke_validate()

    # TODO: look into why this isn't considered fail. something with the way click is
    # running the command... standalone mode, perhaps.
    assert result.success
    assert result.stdout == "invalid"



import json
from pathlib import Path
from unittest.mock import patch

from tests.cmd_util import invoke_new_config
from tests.fixture.config import get_basic_config
from tests.fixture.fsenv import FSEnv
from wap.commands.util import DEFAULT_CONFIG_PATH
from wap.config import Config


def test_new_config() -> None:
    basic_config = get_basic_config()
    with patch(
        "tests.cmd_util.new_config.prompt_for_config",
        return_value=Config.from_python_object(basic_config),
    ):
        result = invoke_new_config()
        assert result.success
        assert json.loads(Path("wap.json").read_text()) == basic_config


def test_new_config_dont_overwrite(fs_env: FSEnv) -> None:
    fs_env.place_file(DEFAULT_CONFIG_PATH)

    result = invoke_new_config(["--config-path", "wap.json"])
    assert isinstance(result.exception, SystemExit)



from typing import Literal
from unittest.mock import DEFAULT, patch

import pytest

from tests.fixture.config import get_basic_config
from tests.fixture.fsenv import FSEnv
from wap import prompt
from wap.wow import FLAVORS, MAINLINE_FLAVOR


@pytest.mark.parametrize("mode", ["new-config", "new-project"])
def test_prompt_for_config(mode: Literal["new-project", "new-config"]) -> None:
    with patch.multiple(
        "wap.prompt",
        get_package_name_new_project=DEFAULT,
        get_package_name_new_config=DEFAULT,
        get_author=DEFAULT,
        get_desc=DEFAULT,
        get_version=DEFAULT,
        get_supported_flavors=DEFAULT,
        get_project_id=DEFAULT,
        get_project_slug=DEFAULT,
        confirm_publish=DEFAULT,
        confirm_all_ok=DEFAULT,
    ) as mocks:
        package_name = "package-name"
        mocks["get_package_name_new_config"].return_value = package_name
        mocks["get_package_name_new_project"].return_value = package_name
        author = "the author"
        mocks["get_author"].return_value = author
        desc = "the desc"
        mocks["get_desc"].return_value = desc
        version = "the version"
        mocks["get_version"].return_value = version
        supported_flavors = [MAINLINE_FLAVOR]
        mocks["get_supported_flavors"].return_value = supported_flavors
        project_id = "1234"
        mocks["get_project_id"].return_value = project_id
        project_slug = "slug"
        mocks["get_project_slug"].return_value = project_slug
        confirm_publish = True
        mocks["confirm_publish"].return_value = confirm_publish
        confirm_all_ok = True
        mocks["confirm_all_ok"].return_value = confirm_all_ok

        config = prompt.prompt_for_config(mode=mode)

        expected_config = {
            "$schema": get_basic_config()["$schema"],
            "name": package_name,
            "version": version,
            "author": author,
            "wowVersions": {
                MAINLINE_FLAVOR.name: MAINLINE_FLAVOR.latest_version.dotted
            },
            "publish": {
                "curseforge": {
                    "projectId": project_id,
                    "slug": project_slug,
                }
            },
            "package": [
                {
                    "path": f"./{package_name}",
                    "toc": {
                        "tags": {
                            "Title": package_name,
                            "Notes": desc,
                        },
                        "files": ["Main.lua"],
                    },
                }
            ],
        }

        assert config.to_python_object() == expected_config


def test_get_package_name_new_project(fs_env: FSEnv) -> None:
    fs_env.place_file("foo")
    with patch("wap.prompt.prompt_ask") as mock:
        mock.side_effect = ["foo", "bar"]
        name = prompt.get_package_name_new_project()
        assert name == "bar"


def test_get_supported_flavors() -> None:
    with patch("wap.prompt.confirm_ask") as mock:
        # simulate answering no to them all first, which is disallowed. then answering
        # yes.
        mock.side_effect = [*[False for _ in FLAVORS], *[True for _ in FLAVORS]]
        flavors = prompt.get_supported_flavors()
        assert set(flavors) == set(FLAVORS)


def test_get_project_id() -> None:
    expected_project_id = "1234"
    with patch("wap.prompt.prompt_ask") as mock:
        mock.side_effect = ["not valid", expected_project_id]
        project_id = prompt.get_project_id()
        assert project_id == expected_project_id


@pytest.mark.parametrize(
    "input, expected_slug",
    [
        (None, None),
        ("slug", "slug"),
    ],
)
def test_get_project_slug(input: str, expected_slug: str) -> None:
    with patch("wap.prompt.prompt_ask") as mock:
        mock.return_value = input
        slug = prompt.get_project_slug()
        assert slug == expected_slug



"""
In testing, we want to see the data submitted to CF when uploading files. It's not
trivial to parse a multipart/form-data request, so this code assists in getting out the
bits of it that we want to verify.
"""

from __future__ import annotations

import json
from io import BytesIO
from typing import Any

import httpx
import multipart  # type: ignore
from attrs import frozen


@frozen(kw_only=True)
class CFUploadRequestContent:
    metadata: Any
    file_content_type: str
    file_name: str
    file_stream: bytes

    @classmethod
    def from_parse(cls, parse: multipart.MultipartParser) -> CFUploadRequestContent:
        metadata = json.loads(parse.get("metadata").raw)  # type: ignore
        file_part = parse.get("file")  # type: ignore
        file_content_type: str = file_part.content_type  # type: ignore
        file_name = file_part.filename  # type: ignore
        file_stream = file_part.raw  # type: ignore
        return CFUploadRequestContent(
            metadata=metadata,
            file_content_type=file_content_type,
            file_name=file_name,  # type: ignore
            file_stream=file_stream,  # type: ignore
        )

    @classmethod
    def from_request(cls, request: httpx.Request) -> CFUploadRequestContent:
        # need to peek to get the boundary
        boundary = request.content.splitlines()[0].removeprefix(b"--")
        return cls.from_parse(
            parse=multipart.MultipartParser(
                stream=BytesIO(request.content), boundary=boundary
            )
        )



from __future__ import annotations

import os
from collections.abc import Iterable, Iterator, Mapping, Sequence
from copy import deepcopy
from pathlib import Path
from typing import Any
from unittest.mock import patch

import pytest
from attrs import frozen
from glom import T, assign, glom  # type: ignore

from tests.cmd_util import invoke_build
from tests.fixture.config import get_basic_config
from tests.fixture.fsenv import FSEnv
from tests.fixture.time import TEST_TIME
from wap import __version__ as wap_version
from wap.exception import (
    ConfigError,
    EncodingError,
    PathExistsError,
    PathMissingError,
    PathTypeError,
    TagError,
)

INSTALLATION_ADDON_DIRS = {
    "mainline": "wow/_retail_/Interface/AddOns",
    "classic": "wow/_classic_/Interface/AddOns",
    "vanilla": "wow/_classic_era_/Interface/AddOns",
}
SUFFIX_INTERFACE_PAIRS = [
    ("_Mainline", "90207"),
    ("_Vanilla", "11403"),
    ("_Classic", "40400"),
    ("", "90207"),
]
PACKAGE_VERSION: str = get_basic_config()["version"]
PACKAGE_NAME = get_basic_config()["name"]


def check_basic_addon(addon_root: Path) -> None:
    config = get_basic_config()
    expected_toc_files = ["Main.lua", "Extra.lua"]
    expected_toc_tags: dict[str, str] = {
        "Title": glom(config, "package.0.toc.tags.Title"),
        "Version": PACKAGE_VERSION,
        # first consult tag Author, then top-level author key
        "Author": (
            glom(config, "package.0.toc.tags.Author", default=None)
            or glom(config, "author", default=None)
        ),
        "Notes": glom(config, "package.0.toc.tags.Notes"),
        "X-BuildDateTime": str(TEST_TIME),
        "X-BuildTool": f"wap {wap_version}",
    }
    for suffix, interface in SUFFIX_INTERFACE_PAIRS:
        path = addon_root / f"{addon_root.name}{suffix}.toc"
        toc = Toc.parse(path)
        assert toc == Toc(
            tags={k: v for k, v in expected_toc_tags.items() if v is not None}
            | {"Interface": interface},
            files=expected_toc_files,
        )

    expected_files = {
        addon_root / "Main.lua",
        addon_root / "Extra.lua",
        addon_root / "LICENSE",
    }
    for expected_file in expected_files:
        assert expected_file.is_file()


@frozen(kw_only=True)
class Toc:
    """
    Rudimentary recreation of toc parsing. just good enough for testing. dont want
    to use the SOT's code to do this.
    """

    tags: Mapping[str, str]
    files: Sequence[str]

    @classmethod
    def parse(cls, path: Path) -> Toc:
        contents = path.read_text()
        tags: dict[str, str] = {}
        files: list[str] = []
        for line in contents.splitlines():
            if line.startswith("##"):
                key, _, value = line.removeprefix("##").partition(":")
                tags[key.strip()] = value.strip()
            elif not line.startswith("#") and line:
                files.append(line.strip())

        return Toc(tags=tags, files=files)


def test_build_normal(fs_env: FSEnv) -> None:
    config = get_basic_config()
    config["package"].append(deepcopy(config["package"][0]))
    assign(config, "package.1.path", "./Addon2")
    fs_env.write_config(config)

    fs_env.place_addon("basic")
    fs_env.place_addon("basic", "Addon2")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert result.success

    check_basic_addon(Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon"))
    check_basic_addon(Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon2"))


def test_build_dupe_addon_path(fs_env: FSEnv) -> None:
    config = get_basic_config()
    config["package"].append(config["package"][0])
    fs_env.write_config(config)

    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert isinstance(result.exception, ConfigError)


def test_build_different_output_path(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    output_path = "output"

    result = invoke_build(["--output-path", output_path])

    assert result.success

    check_basic_addon(Path(f"{output_path}/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon"))


def test_build_different_config_path(fs_env: FSEnv) -> None:
    config_path = "config.json"
    fs_env.write_config(get_basic_config(), config_path)
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build(["--config-path", config_path])

    assert result.success

    check_basic_addon(Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon"))


def test_build_config_discovery(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    os.chdir(Path("Addon"))

    result = invoke_build()

    assert result.success

    check_basic_addon(Path(f"../dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon"))


def test_build_config_discovery_fail(fs_env: FSEnv) -> None:
    result = invoke_build()

    assert isinstance(result.exception, SystemExit)


@pytest.mark.parametrize("top_level_author", [True, False])
@pytest.mark.parametrize("toc_author", [True, False])
def test_build_author_fallback(
    fs_env: FSEnv, top_level_author: bool, toc_author: bool
) -> None:
    config = get_basic_config()

    top_level_author_name = "top-level"
    toc_author_name = "toc"
    if top_level_author:
        assign(config, "author", top_level_author_name)
    if toc_author:
        assign(config, "author", toc_author_name)

    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert result.success

    check_basic_addon(Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon"))


def test_build_missing_config() -> None:
    result = invoke_build()

    # i wish a more descriptive exception was thrown than SystemExit, but we have no
    # choice because click controls this raise
    assert isinstance(result.exception, SystemExit)


def test_build_config_is_not_file(fs_env: FSEnv) -> None:
    fs_env.place_dir("wap.json")

    result = invoke_build()

    assert isinstance(result.exception, SystemExit)


def test_build_missing_addon(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert isinstance(result.exception, PathTypeError)


def test_build_addon_path_is_not_dir(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_file("Addon")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert isinstance(result.exception, PathTypeError)


@pytest.mark.parametrize(
    "link_arg",
    [
        "mainline",
        "classic",
        "vanilla",
    ],
)
def test_build_with_linking(fs_env: FSEnv, link_arg: str) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    args: list[str] = []
    for flavor, path in INSTALLATION_ADDON_DIRS.items():
        args.extend((f"--{flavor}-addons-path", path))
        fs_env.place_dir(path, parents=True, exist_ok=True)

    args.extend(["--link", link_arg])

    result = invoke_build(args)

    assert result.success
    assert (
        Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon").resolve()
        == (Path(INSTALLATION_ADDON_DIRS[link_arg]) / "Addon").resolve()
    )
    assert "Linked" in result.stderr


@pytest.mark.parametrize(
    "link_arg",
    [
        "mainline",
        "classic",
        "vanilla",
    ],
)
def test_build_link_already_linked(fs_env: FSEnv, link_arg: str) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    args: list[str] = []
    for flavor, path in INSTALLATION_ADDON_DIRS.items():
        args.extend((f"--{flavor}-addons-path", path))
        fs_env.place_dir(path, parents=True, exist_ok=True)

    # create initial fake output dir so that linking works
    build_dir = fs_env.place_dir(
        f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon", parents=True
    )
    (Path(INSTALLATION_ADDON_DIRS[link_arg]) / "Addon").symlink_to(
        build_dir, target_is_directory=True
    )

    args.extend(["--link", link_arg])

    result = invoke_build(args)

    assert result.success
    assert (
        Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon").resolve()
        == (Path(INSTALLATION_ADDON_DIRS[link_arg]) / "Addon").resolve()
    )


@pytest.mark.parametrize(
    "link_arg",
    [
        "mainline",
        "classic",
        "vanilla",
    ],
)
def test_build_link_path_exists(fs_env: FSEnv, link_arg: str) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    args: list[str] = []
    for flavor, path in INSTALLATION_ADDON_DIRS.items():
        args.extend((f"--{flavor}-addons-path", path))
        fs_env.place_dir(path, parents=True, exist_ok=True)

    # create a dir in the spot where the link would be made
    fs_env.place_dir(str(Path(INSTALLATION_ADDON_DIRS[link_arg]) / "Addon"))

    args.extend(["--link", link_arg])

    result = invoke_build(args)

    assert isinstance(result.exception, PathExistsError)


@pytest.mark.parametrize(
    "link_arg",
    [
        "mainline",
        "classic",
        "vanilla",
    ],
)
def test_build_link_path_exists_force(fs_env: FSEnv, link_arg: str) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    args: list[str] = []
    for flavor, path in INSTALLATION_ADDON_DIRS.items():
        args.extend((f"--{flavor}-addons-path", path))
        fs_env.place_dir(path, parents=True, exist_ok=True)

    # create a dir in the spot where the link would be made
    fs_env.place_dir(str(Path(INSTALLATION_ADDON_DIRS[link_arg]) / "Addon"))

    args.extend(["--link", link_arg, "--link-force"])

    result = invoke_build(args)

    assert result.success
    assert (
        Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon").resolve()
        == (Path(INSTALLATION_ADDON_DIRS[link_arg]) / "Addon").resolve()
    )


@pytest.mark.parametrize("add_auto", [True, False])
@pytest.mark.parametrize(
    "exist_flavor",
    [
        "mainline",
        "classic",
        "vanilla",
    ],
)
def test_build_with_auto_linking(
    fs_env: FSEnv, add_auto: bool, exist_flavor: str
) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    args: list[str] = []
    for flavor, path in INSTALLATION_ADDON_DIRS.items():
        args.extend((f"--{flavor}-addons-path", path))
        if exist_flavor == flavor:
            fs_env.place_dir(path, parents=True, exist_ok=True)

    args.append("--link")
    if add_auto:
        args.append("auto")

    result = invoke_build(args)

    assert result.success
    assert (
        Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon").resolve()
        == (Path(INSTALLATION_ADDON_DIRS[exist_flavor]) / "Addon").resolve()
    )
    for flavor, path in INSTALLATION_ADDON_DIRS.items():
        if exist_flavor != flavor:
            assert not (Path(path) / "Addon").exists()


@pytest.mark.parametrize(
    "other_flavor",
    [
        "mainline",
        "classic",
        "vanilla",
    ],
)
def test_build_link_auto_and_other(fs_env: FSEnv, other_flavor: str) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build(["--link", "auto", "--link", other_flavor])

    assert isinstance(result.exception, SystemExit)


def test_build_non_unicode_config(fs_env: FSEnv) -> None:
    config_path = fs_env.write_config(get_basic_config())
    config_path.write_text(
        config_path.read_text().replace("Package", "包裹"), encoding="big5"
    )
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert isinstance(result.exception, EncodingError)


def test_build_overwriting_src_include(fs_env: FSEnv) -> None:
    fs_env.write_config(assign(get_basic_config(), "package.0.include", ["./Main.lua"]))
    fs_env.place_addon("basic")
    fs_env.place_file("Main.lua")

    result = invoke_build()

    assert result.success
    assert "already exists" in result.stderr


def test_build_overwriting_toc_include(fs_env: FSEnv) -> None:
    fs_env.write_config(
        assign(get_basic_config(), "package.0.include", ["./Addon.toc"])
    )
    fs_env.place_addon("basic")
    fs_env.place_file("Addon.toc")

    result = invoke_build()

    assert result.success
    assert "already exists" in result.stderr


def test_build_addon_dir_exists_as_file(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon", parents=True)

    result = invoke_build()

    assert isinstance(result.exception, PathExistsError)


@pytest.mark.parametrize(
    "clean",
    [
        True,
        False,
    ],
)
def test_build_cleaning(fs_env: FSEnv, clean: bool) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    old = fs_env.place_file(
        f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon/old", parents=True
    )

    args = ["--clean"] if clean else []
    result = invoke_build(args)

    assert result.success
    assert old.exists() != clean


def test_build_include_missing(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")

    result = invoke_build()

    assert result.success
    assert "none were found" in result.stderr


def test_build_include_dir(fs_env: FSEnv) -> None:
    fs_env.write_config(
        assign(get_basic_config(), "package.0.include", ["./a-directory"])
    )
    fs_env.place_addon("basic")
    fs_env.place_dir("a-directory")

    result = invoke_build()

    assert result.success
    assert Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon/a-directory").is_dir()


def test_build_toc_file_missing(fs_env: FSEnv) -> None:
    config = get_basic_config()
    glom(config, ("package.0.toc.files", T.append("NotExist.lua")))
    fs_env.write_config(config)
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert isinstance(result.exception, PathMissingError)


@pytest.mark.parametrize(
    "toc_suffix",
    [
        "",
        "_Classic",
        "_Vanilla",
    ],
)
def test_build_gen_toc_overwrites(fs_env: FSEnv, toc_suffix: str) -> None:
    fs_env.write_config(get_basic_config())
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    fs_env.place_file(f"Addon/Addon{toc_suffix}.toc")

    result = invoke_build()

    assert result.success
    assert "already exists" in result.stderr


def test_build_include_dir_overwrite_file(fs_env: FSEnv) -> None:
    fs_env.write_config(
        assign(get_basic_config(), "package.0.include", ["./a", "./b/a"])
    )
    fs_env.place_addon("basic")
    fs_env.place_file("a")
    fs_env.place_dir("b/a", parents=True)

    result = invoke_build()

    assert isinstance(result.exception, PathExistsError)


def test_build_include_file_overwrite_file(fs_env: FSEnv) -> None:
    fs_env.write_config(
        assign(get_basic_config(), "package.0.include", ["./a", "./b/a"])
    )
    fs_env.place_addon("basic")
    fs_env.place_file("a", text="first")
    fs_env.place_file("b/a", parents=True, text="last")

    result = invoke_build()

    assert result.success
    assert Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon/a").read_text() == "last"


def test_build_include_dir_overwrite_dir(fs_env: FSEnv) -> None:
    fs_env.write_config(
        assign(get_basic_config(), "package.0.include", ["./a", "./b/a"])
    )
    fs_env.place_addon("basic")
    fs_env.place_dir("a")
    fs_env.place_dir("b/a", parents=True)

    result = invoke_build()

    assert result.success
    assert Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon/a").is_dir()


def test_build_include_file_overwrite_dir(fs_env: FSEnv) -> None:
    fs_env.write_config(
        assign(get_basic_config(), "package.0.include", ["./a", "./b/a"])
    )
    fs_env.place_addon("basic")
    fs_env.place_dir("a")
    fs_env.place_file("b/a", parents=True)

    result = invoke_build()

    assert isinstance(result.exception, PathExistsError)


def test_build_toc_included(fs_env: FSEnv) -> None:
    fs_env.write_config(
        assign(get_basic_config(), "package.0.include", ["./Addon.toc"])
    )
    fs_env.place_addon("basic")
    fs_env.place_dir("Addon.toc")

    result = invoke_build()

    assert isinstance(result.exception, PathExistsError)


@pytest.mark.parametrize("load_on_demand,expected", [(True, "1"), (False, "0")])
def test_build_toc_load_on_demand(
    fs_env: FSEnv, load_on_demand: bool, expected: str
) -> None:
    fs_env.write_config(
        assign(get_basic_config(), "package.0.toc.tags.LoadOnDemand", load_on_demand)
    )
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert result.success

    for suffix, _ in SUFFIX_INTERFACE_PAIRS:
        toc_path = Path(
            f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon/Addon{suffix}.toc"
        )
        toc = Toc.parse(toc_path)
        assert {("LoadOnDemand", expected)} <= toc.tags.items()


@pytest.mark.parametrize(
    "default_state,expected", [(True, "enabled"), (False, "disabled")]
)
def test_build_toc_default_state(
    fs_env: FSEnv, default_state: bool, expected: str
) -> None:
    fs_env.write_config(
        assign(get_basic_config(), "package.0.toc.tags.DefaultState", default_state)
    )
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert result.success

    for suffix, _ in SUFFIX_INTERFACE_PAIRS:
        toc_path = Path(
            f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon/Addon{suffix}.toc"
        )
        toc = Toc.parse(toc_path)
        assert {("DefaultState", expected)} <= toc.tags.items()


@pytest.mark.parametrize(
    "tag_name",
    [
        "Dependencies",
        "OptionalDeps",
        "LoadWith",
        "LoadManagers",
        "SavedVariables",
        "SavedVariablesPerCharacter",
    ],
)
@pytest.mark.parametrize(
    "tag_value,expected", [([], ""), (["a"], "a"), (["a", "b", "c"], "a, b, c")]
)
def test_build_toc_list_type_tags(
    fs_env: FSEnv, tag_name: str, tag_value: list[str], expected: str
) -> None:
    fs_env.write_config(
        assign(get_basic_config(), f"package.0.toc.tags.{tag_name}", tag_value)
    )
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert result.success

    for suffix, _ in SUFFIX_INTERFACE_PAIRS:
        toc_path = Path(
            f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon/Addon{suffix}.toc"
        )
        toc = Toc.parse(toc_path)
        assert {(tag_name, expected)} <= toc.tags.items()


@pytest.mark.parametrize("bad_tag", ["foo bar", "foo\nbar", "foo:bar"])
def test_build_toc_bad_tags(fs_env: FSEnv, bad_tag: str) -> None:
    config = get_basic_config()
    config["package"][0]["toc"]["tags"][bad_tag] = "value"
    fs_env.write_config(config)
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    result = invoke_build()

    assert isinstance(result.exception, TagError)


def test_build_watch_edit_source_file(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())
    addon_path = fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    changing_file_path = addon_path / "Main.lua"
    original_text = "original"
    new_text = "updated"
    assert original_text != new_text

    changing_file_path.write_text(original_text)

    def edit_file(*args: Any, **kwargs: Any) -> Iterator[Iterable[Path]]:
        changing_file_path.write_text(new_text)
        yield {changing_file_path.resolve()}

    with patch("tests.cmd_util.build.watch_paths", side_effect=edit_file):
        result = invoke_build(["--watch"])

    assert result.success

    new_output_file_path = Path(f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon/Main.lua")
    assert new_output_file_path.read_text() == new_text


def test_build_watch_new_source_file(fs_env: FSEnv) -> None:
    fs_env.write_config(get_basic_config())
    addon_path = fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    # create a new file in a directory
    new_file_path = addon_path / "Dir" / "New.lua"
    new_text = "new"

    def create_file(*args: Any, **kwargs: Any) -> Iterator[Iterable[Path]]:
        new_file_path.parent.mkdir()
        new_file_path.write_text(new_text)
        yield {new_file_path.resolve()}

    with patch("tests.cmd_util.build.watch_paths", side_effect=create_file):
        result = invoke_build(["--watch"])

    assert result.success

    new_output_file_path = Path(
        f"dist/{PACKAGE_NAME}-{PACKAGE_VERSION}/Addon/Dir/New.lua"
    )
    assert new_output_file_path.read_text() == new_text


def test_build_watch_config_file(fs_env: FSEnv) -> None:
    config = get_basic_config()
    fs_env.write_config(config)
    fs_env.place_addon("basic")
    fs_env.place_file("LICENSE")

    new_version = "9.9.9"
    new_text = "new"
    fs_env.place_file("New.txt", text=new_text)

    def edit_config(*args: Any, **kwargs: Any) -> Iterator[Iterable[Path]]:
        # do two arbitrary things that we can later test:
        # 1. add another include (so it will show up in the output dir)
        # 2. change the package version (so the output dir path changes)
        config["package"][0]["include"].append("./New.txt")
        config["version"] = new_version
        config_path = fs_env.write_config(config)
        yield {config_path.resolve()}

    with patch("tests.cmd_util.build.watch_paths", side_effect=edit_config):
        result = invoke_build(["--watch"])

    assert result.success

    assert (
        Path(f"dist/{PACKAGE_NAME}-{new_version}/Addon/New.txt").read_text() == new_text
    )



import json
from pathlib import Path
from unittest.mock import patch

from tests.cmd_util import invoke_new_project
from tests.fixture.config import get_basic_config
from tests.fixture.fsenv import FSEnv
from wap.config import Config


def test_new_project(fs_env: FSEnv) -> None:
    basic_config = get_basic_config()
    with patch(
        "tests.cmd_util.new_project.prompt_for_config",
        return_value=Config.from_python_object(basic_config),
    ):
        result = invoke_new_project()
        assert result.success

        project_root = Path("Package")
        assert json.loads((project_root / "wap.json").read_text()) == basic_config
        assert (project_root / "README.md").is_file()
        assert (project_root / "Package/Main.lua").is_file()



"""
A mock version of the CurseForge API
"""

import random
import re
import uuid
from collections.abc import Callable

from httpx import Request, Response
from respx.router import MockRouter

from wap.curseforge import CurseForgeAPI

# simulate actual authentication by declaring the only working token -- all others fail.
# callers will need to use this.
CURSEFORGE_TOKEN = str(uuid.uuid4())


def _check_auth(fn: Callable[[Request], Response]) -> Callable[[Request], Response]:
    def check(request: Request, **kwargs: str) -> Response:
        api_token = request.headers[CurseForgeAPI.TOKEN_HEADER_NAME]
        if not api_token:
            return Response(
                401,
                json={
                    "errorCode": 401,
                    "errorMessage": (
                        "You must provide an API token using the `X-Api-Token` header, "
                        "the `token` query string parameter, your email address and an "
                        "API token using HTTP basic authentication."
                    ),
                },
            )
        if not re.fullmatch(
            r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
            api_token,
            flags=re.IGNORECASE,
        ):
            return Response(
                400,
                json={
                    "errorCode": 3,
                    "errorMessage": (
                        f"API token is malformed. Token provided: {api_token}."
                    ),
                },
            )
        if api_token != CURSEFORGE_TOKEN:
            return Response(
                403,
                json={"errorCode": 403, "errorMessage": "Invalid API token provided."},
            )
        return fn(request, **kwargs)

    return check


@_check_auth
def _version_response(_: Request, **kwargs: str) -> Response:
    # kwargs allows respx to pass in named regex groups as key-value pairs if they are
    # specified in the route
    return Response(
        200,
        json=[
            {"name": "9.2.7", "id": 1000},
            {"name": "9.2.7", "id": 1001},  # this one has higher id though
            {"name": "4.4.0", "id": 1002},
            {"name": "1.14.3", "id": 1003},
        ],
    )


@_check_auth
def _upload_response(_: Request, **kwargs: str) -> Response:
    return Response(200, json={"id": random.randint(1_000, 9_999)})


def setup_mock_cf_api(respx_mock: MockRouter) -> MockRouter:
    respx_mock.get(
        url="https://wow.curseforge.com/api/game/versions", name="versions"
    ).mock(side_effect=_version_response)
    respx_mock.post(
        url__regex=(
            r"https://wow.curseforge.com/api/projects/(?P<project_id>\d+)/upload-file"
        ),
        name="upload-file",
    ).mock(side_effect=_upload_response)
    return respx_mock



from __future__ import annotations

import re
from collections.abc import Sequence

from attr import frozen
from click import BaseCommand
from click.testing import CliRunner, Result

# we need to patch some functions in these module, so we gotta import the module itself
# -- can't import specific things from the module because then the references wouldn't
# be patchable.
from wap.commands import base, build, new_config, new_project, publish, validate

_WHITESPACE_PATTERN = r"\s+"


@frozen(kw_only=True)
class RunResult:  # yeesh, this name is derivative
    """
    The result of an invoked Click command with some quality-of-life improvements.
    """

    stdout_raw: str
    stderr_raw: str
    exit_code: int
    exception: BaseException | None

    @classmethod
    def from_result(cls, result: Result) -> RunResult:
        return RunResult(
            stdout_raw=result.stdout,
            stderr_raw=result.stderr,
            exit_code=result.exit_code,
            exception=result.exception,
        )

    @property
    def stdout(self) -> str:
        # ensures there's no line breaks to mess up __contains__ calls
        return re.sub(_WHITESPACE_PATTERN, " ", self.stdout_raw).strip()

    @property
    def stderr(self) -> str:
        return re.sub(_WHITESPACE_PATTERN, " ", self.stderr_raw).strip()

    @property
    def success(self) -> bool:
        return self.exit_code == 0


def invoke(command: BaseCommand, args: Sequence[str] | None = None) -> RunResult:
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(command, args or [])
    return RunResult.from_result(result)


def invoke_build(args: Sequence[str] | None = None) -> RunResult:
    return invoke(build.build, args)


def invoke_validate(args: Sequence[str] | None = None) -> RunResult:
    return invoke(validate.validate, args)


def invoke_help(args: Sequence[str] | None = None) -> RunResult:
    return invoke(base.help_command, args)


def invoke_new_config(args: Sequence[str] | None = None) -> RunResult:
    return invoke(new_config.new_config, args)


def invoke_new_project(args: Sequence[str] | None = None) -> RunResult:
    return invoke(new_project.new_project, args)


def invoke_publish(args: Sequence[str] | None = None) -> RunResult:
    return invoke(publish.publish, args)
